#include "lib_nxwms.h"

int main (int ac, char *av[]) {
  int pid;int mid;
  pid=3;mid=1;
  if ((ac != 3) && (ac != 1)){
    fprintf(stderr,"%s product_id_to_check modlue_id_to_check\n",av[0]);
    return 1;
  }
  if (ac == 3){
    pid=atoi(av[1]);
    mid=atoi(av[2]);
  }
  int fd = open_nxwms(NULL);
  if(set_measurefrequency(fd, pid, mid,500,0) == 1){
    fprintf(stdout,"ok pid %d mid %d\n",pid,mid);
  }
  close_nxwms(fd);
  printf("OK?\n");
  return 0;
}
